@@ -167,6 +167,86 @@ static bool CheckGPGV(std::unordered_map<std::string, std::forward_list<std::str
    return not checkedCommands[gpgv].empty();
 }
 
+/// Verifies a file containing a detached signature has the right format
+/// @return 0 if succesful, or an exit code for ExecGPGV otherwise.
+static int VerifyDetachedSignatureFile(std::string FileGPG, int fd[2], int statusfd = -1)
+{
+   auto detached = make_unique_FILE(FileGPG, "r");
+   if (detached.get() == nullptr)
+      return _error->Error("Detached signature file '%s' could not be opened", FileGPG.c_str()), 111;
+
+   LineBuffer buf;
+   bool open_signature = false;
+   bool found_badcontent = false;
+   size_t found_signatures = 0;
+   while (buf.readFrom(detached.get(), FileGPG, true))
+   {
+      if (open_signature)
+      {
+	 if (buf == "-----END PGP SIGNATURE-----")
+	    open_signature = false;
+	 else if (buf.starts_with("-"))
+	 {
+	    // the used Radix-64 is not using dash for any value, so a valid line can't
+	    // start with one. Header keys could, but no existent one does and seems unlikely.
+	    // Instead it smells a lot like a header the parser didn't recognize.
+	    _error->Error("Detached signature file '%s' contains unexpected line starting with a dash", FileGPG.c_str());
+	    return 112;
+	 }
+      }
+      else // if (not open_signature)
+      {
+	 if (buf == "-----BEGIN PGP SIGNATURE-----")
+	 {
+	    open_signature = true;
+	    ++found_signatures;
+	    if (found_badcontent)
+	       break;
+	 }
+	 else
+	 {
+	    found_badcontent = true;
+	    if (found_signatures != 0)
+	       break;
+	 }
+      }
+   }
+   if (found_signatures == 0 && statusfd != -1)
+   {
+      auto const errtag = "[GNUPG:] NODATA\n";
+      FileFd::Write(fd[1], errtag, strlen(errtag));
+      // guess if this is a binary signature, we never officially supported them,
+      // but silently accepted them via passing them unchecked to gpgv
+      if (found_badcontent)
+      {
+	 rewind(detached.get());
+	 auto ptag = fgetc(detached.get());
+	 // §4.2 says that the first bit is always set and gpg seems to generate
+	 // only old format which is indicated by the second bit not set
+	 if (ptag != EOF && (ptag & 0x80) != 0 && (ptag & 0x40) == 0)
+	 {
+	    _error->Error("Detached signature file '%s' is in unsupported binary format", FileGPG.c_str());
+	    return 112;
+	 }
+      }
+      // This is not an attack attempt but a file even gpgv would complain about
+      // likely the result of a paywall which is covered by the gpgv method
+      return 113;
+   }
+   else if (found_badcontent)
+   {
+      _error->Error("Detached signature file '%s' contains lines not belonging to a signature", FileGPG.c_str());
+      return 112;
+   }
+   if (open_signature)
+   {
+      _error->Error("Detached signature file '%s' contains unclosed signatures", FileGPG.c_str());
+      return 112;
+   }
+
+   return 0;
+}
+
 std::pair<std::string, std::forward_list<std::string>> APT::Internal::FindGPGV(bool Debug)
 {
    static thread_local std::unordered_map<std::string, std::forward_list<std::string>> checkedCommands;
@@ -222,13 +302,17 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    Args.push_back(gpgv);
    Args.push_back("--ignore-time-conflict");
 
-   auto dearmorKeyOrCheckFormat = [&](std::string const &k) -> std::string
+   FileFd mergedFd;
+   if (GetTempFile("apt.XXXXXX.gpg", false, &mergedFd) == nullptr)
+      local_exit(EINTERNAL);
+   local_exit.files.push_back(mergedFd.Name());
+
+   auto dearmorKeyOrCheckFormat = [&](std::string const &k)
    {
       FileFd keyFd(k, FileFd::ReadOnly);
       if (not keyFd.IsOpen())
       {
-	 apt_warning(std::cerr, statusfd, fd, "The key(s) in the keyring %s are ignored as the file is not readable by user executing apt-key.\n", k.c_str());
-	 return "";
+	 apt_warning(std::cerr, statusfd, fd, "The key(s) in the keyring %s are ignored as the file is not readable by user executing gpgv.\n", k.c_str());
       }
       else if (APT::String::Endswith(k, ".asc"))
       {
@@ -249,13 +333,10 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
 	 if (state != 3)
 	    goto err;
 
-	 FileFd dearmoredFd;
-	 if (GetTempFile("apt.XXXXXX.gpg", false, &dearmoredFd) == nullptr)
-	    local_exit(EINTERNAL);
 	 if (auto decoded = Base64Decode(b64msg); not decoded.empty())
-	    dearmoredFd.Write(decoded.data(), decoded.size());
-	 local_exit.files.push_back(dearmoredFd.Name());
-	 return dearmoredFd.Name();
+	    if (not mergedFd.Write(decoded.data(), decoded.size()))
+	       local_exit(EINTERNAL);
+	 return;
       }
       else
       {
@@ -266,25 +347,29 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
 	 // 0x98 -- old-format OpenPGP public key packet, up to 255 octets
 	 // 0x99 -- old-format OpenPGP public key packet, 256-65535 octets
 	 // 0xc6 -- new-format OpenPGP public key packet, any length
-	 if (c == 0x98 || c == 0x99 || c == 0xc6)
-	    return k;
+	 if (c != 0x98 && c != 0x99 && c != 0xc6)
+	    goto err;
+
+	 if (not mergedFd.Write(&c, sizeof(c)))
+	    local_exit(EINTERNAL);
+
+	 if (not CopyFile(keyFd, mergedFd))
+	    local_exit(EINTERNAL);
+
+	 return;
       }
    err:
       apt_warning(std::cerr, statusfd, fd, "The key(s) in the keyring %s are ignored as the file has an unsupported filetype.", k.c_str());
-      return "";
    };
    auto maybeAddKeyring = [&](std::string const &k)
    {
       if (struct stat st; stat(k.c_str(), &st) != 0 || st.st_size == 0)
 	 return;
-      if (auto cleanKey = dearmorKeyOrCheckFormat(k); not cleanKey.empty())
-      {
-	 Args.push_back("--keyring");
-	 Args.push_back(cleanKey);
-      }
+      dearmorKeyOrCheckFormat(k);
       return;
    };
 
+   bool FoundKeyring = false;
    for (auto const &k : KeyFiles)
    {
       if (unlikely(k.empty()))
@@ -295,6 +380,7 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
 	    std::clog << "Trying Signed-By: " << k << std::endl;
 
 	 maybeAddKeyring(k);
+	 FoundKeyring = true;
       }
       else
       {
@@ -304,7 +390,7 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    }
 
    std::vector<std::string> Parts;
-   if (std::find(Args.begin(), Args.end(), "--keyring") == Args.end())
+   if (not FoundKeyring)
    {
       Parts = GetListOfFilesInDir(_config->FindDir("Dir::Etc::TrustedParts"), std::vector<std::string>{"gpg", "asc"}, true);
       if (char *env = getenv("APT_KEY_NO_LEGACY_KEYRING"); env == nullptr || not StringToBool(env, false))
@@ -318,11 +404,8 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    }
 
    // If we do not give it any keyring, gpgv shouts keydb errors at us
-   if (std::find(Args.begin(), Args.end(), "--keyring") == Args.end())
-   {
-      Args.push_back("--keyring");
-      Args.push_back("/dev/null");
-   }
+   Args.push_back("--keyring");
+   Args.push_back(mergedFd.Name());
 
    char statusfdstr[10];
    if (statusfd != -1)
@@ -355,81 +438,18 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
 
    if (releaseSignature == DETACHED)
    {
-      auto detached = make_unique_FILE(FileGPG, "r");
-      if (detached.get() == nullptr)
+      // Collect the error and return it via apt_error()
+      _error->PushToStack();
+      auto exitCode = VerifyDetachedSignatureFile(FileGPG, fd, statusfd);
+      std::string msg;
+      _error->PopMessage(msg);
+      _error->RevertToStack();
+      if (exitCode != 0)
       {
-	 apt_error(std::cerr, statusfd, fd, "Detached signature file '%s' could not be opened", FileGPG.c_str());
-	 local_exit(EINTERNAL);
+	 if (not msg.empty())
+	    apt_error(std::cerr, statusfd, fd, "%s", msg.c_str());
+	 local_exit(exitCode);
       }
-      LineBuffer buf;
-      bool open_signature = false;
-      bool found_badcontent = false;
-      size_t found_signatures = 0;
-      while (buf.readFrom(detached.get(), FileGPG, true))
-      {
-	 if (open_signature)
-	 {
-	    if (buf == "-----END PGP SIGNATURE-----")
-	       open_signature = false;
-	    else if (buf.starts_with("-"))
-	    {
-	       // the used Radix-64 is not using dash for any value, so a valid line can't
-	       // start with one. Header keys could, but no existent one does and seems unlikely.
-	       // Instead it smells a lot like a header the parser didn't recognize.
-	       apt_error(std::cerr, statusfd, fd, "Detached signature file '%s' contains unexpected line starting with a dash", FileGPG.c_str());
-	       local_exit(112);
-	    }
-	 }
-	 else //if (not open_signature)
-	 {
-	    if (buf == "-----BEGIN PGP SIGNATURE-----")
-	    {
-	       open_signature = true;
-	       ++found_signatures;
-	       if (found_badcontent)
-		  break;
-	    }
-	    else
-	    {
-	       found_badcontent = true;
-	       if (found_signatures != 0)
-		  break;
-	    }
-	 }
-      }
-      if (found_signatures == 0 && statusfd != -1)
-      {
-	 auto const errtag = "[GNUPG:] NODATA\n";
-	 FileFd::Write(fd[1], errtag, strlen(errtag));
-	 // guess if this is a binary signature, we never officially supported them,
-	 // but silently accepted them via passing them unchecked to gpgv
-	 if (found_badcontent)
-	 {
-	    rewind(detached.get());
-	    auto ptag = fgetc(detached.get());
-	    // §4.2 says that the first bit is always set and gpg seems to generate
-	    // only old format which is indicated by the second bit not set
-	    if (ptag != EOF && (ptag & 0x80) != 0 && (ptag & 0x40) == 0)
-	    {
-	       apt_error(std::cerr, statusfd, fd, "Detached signature file '%s' is in unsupported binary format", FileGPG.c_str());
-	       local_exit(112);
-	    }
-	 }
-	 // This is not an attack attempt but a file even gpgv would complain about
-	 // likely the result of a paywall which is covered by the gpgv method
-	 local_exit(113);
-      }
-      else if (found_badcontent)
-      {
-	 apt_error(std::cerr, statusfd, fd, "Detached signature file '%s' contains lines not belonging to a signature", FileGPG.c_str());
-	 local_exit(112);
-      }
-      if (open_signature)
-      {
-	 apt_error(std::cerr, statusfd, fd, "Detached signature file '%s' contains unclosed signatures", FileGPG.c_str());
-	 local_exit(112);
-      }
-
       Args.push_back(FileGPG);
       Args.push_back(File);
    }
@@ -509,14 +529,14 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    {
       if (errno == EINTR)
 	 continue;
-      apt_error(std::cerr, statusfd, fd, _("Waited for %s but it wasn't there"), "apt-key");
+      apt_error(std::cerr, statusfd, fd, _("Waited for %s but it wasn't there"), gpgv.c_str());
       local_exit(EINTERNAL);
    }
 
    // check if it exit'ed normally …
    if (not WIFEXITED(Status))
    {
-      apt_error(std::cerr, statusfd, fd, _("Sub-process %s exited unexpectedly"), "apt-key");
+      apt_error(std::cerr, statusfd, fd, _("Sub-process %s exited unexpectedly"), gpgv.c_str());
       local_exit(EINTERNAL);
    }
 
@@ -524,7 +544,7 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    if (WEXITSTATUS(Status) != 0)
    {
       // we forward the statuscode, so don't generate a message on the fd in this case
-      apt_error(std::cerr, -1, fd, _("Sub-process %s returned an error code (%u)"), "apt-key", WEXITSTATUS(Status));
+      apt_error(std::cerr, -1, fd, _("Sub-process %s returned an error code (%u)"), gpgv.c_str(), WEXITSTATUS(Status));
       local_exit(WEXITSTATUS(Status));
    }
 
